#include "gc.h"
#include "debug.h"


/* This GC dumb:
 * it copies everything that's pointed to,
 * hence what is collected is head garbage
 * if a list '(a b c d e) is unreferenced,
 * then 5 flips are needed to completely collect it,
 * but we needn't worry as long as k is such that after
 * a memory flip, the total memory actually used verifies:
 *
 * if x is the used/total memory ratio,  k >= 1/(1-x)   x <= 1-1/k
 *
 * The problem is that hbaker's algorithm does not
 * allow to detect that, and hence, because it flips
 * only when memory is full, and then it cannot evaluate x,
 *
 * A patch to the method would be to flip as soon
 * so the only conservative approach would be x=1-A/N
 * (leaving A cells among N), hence k=N/A.
 * We see that
 *
 * Also, though it may be space efficient with large enough k,
 * there is no recovery routine possible in case too many CONS
 * were done during a flip, and the algorithm provides no way to detect
 * that before it's too late.
 *
 * An adaptative algorithm might flip as soon as
 * scanptr reaches bottom (or run until x>3/4 if bottom is too low,
 * to ensure a min k of 5)
 * so as to guarantee not too high a k for the next flip.
 * if not enough memory was freed during the flip,
 * then k will have to be increased according to the x being measured.
 * Beware that too low a k may mean that next flip will have a much much
 * higher k, so we may like not have k go below the average measured k
 * (the equilibrium value that hbaker@netcom.com tells us about).
 * For instance, we could use (measured_k+previous_k+1)/2
 * if previous_k>measured_k, and max(5,k) in all cases.
 * 
 *
 * If we use such an adaptative k, then we're sure that memory can
 * not overflow without a clear warning beforehand.
 * We can trap the overflow by detecting k>kmax,
 * where kmax is the max limit tolerated by real-time behavior.
 * By computing k with N'<N, we can even reserve the space N-N'
 * for memory overflow exception handling!
 * 
 */

/*
 * WARNING!
 * The GC assumes all registers were updated before next flip;
 * surely flip() should update all registers/global variables immediately!
 */


int k = 5;              /* Ratio of GC's per CONS. Higher = more, slower    */
/* k should be adjusted according to the */

inline void GC_increment() {
  int i;

  DBG_printf("GC_increment: %p < %p?\n", scanptr, bottom);
  for(i = 0; (i < k) && (scanptr < bottom); i++) {
    car(scanptr);
    cdr(scanptr);
    scanptr += 2;
  }
  DBG_printf("GC_increment: done %d.\n", i);
  
}

void flip(void) {  /* 'flip' the spaces. No error-checking, quick and dirty */

  DBG_printf("flip()\n");

  SWAP( tospace, fromspace );
  bottom      = tospace;
  top         = tospace + space_size - 2;
  scanptr     = tospace;

  DBG(  dump_ptrs()  );
}


inline int flipping(int k) {
  int bool = (bottom == top) ;

  if(bool) {
    if(scanptr < bottom) error("scanptr < bottom???\n");
    flip();
  }
  return bool ;
}



inline int in_fromspace(PTR p) {
  int d ;
  DBG_printf("in_fromspace(%p) fromspace=%p-%p\n",
		p, fromspace, fromspace+space_size-1);
  d = p-fromspace ;
  return (d>=0) && (d<space_size) ;
}

inline int in_tospace(PTR p) {
  int d = p-tospace ;
  return (d>=0) && (d<space_size) ;
}


inline PTR move_ptr(PTR p) {

  DBG_printf("move_ptr(%p)\n",p);

  if(!in_fromspace(p)) return p;
  else {

    DBG_printf("We're in fromspace, move me\n");
    DBG_printf("if(!in_tospace(%p)) __CAR(p) = copy(%p);\n",M2P(__CAR(p)),p);

    if( atom(__CAR(p)) )
      	__RPLACA(p,MEM_of_PTR(copy_cell(p))) ;

    DBG(  dump_ptrs()  );

    return PTR_of_MEM(__CAR(p));
  }
}

inline MEM move_mem(MEM x) {
  if (atom(x)) return x;
  else return MEM_of_PTR(move_ptr(PTR_of_MEM(x)));
}

inline PTR copy_cell(PTR p) {
  PTR newcell;

  DBG_printf("copy_cell(%p)\n",p);

  if(bottom >= top) error("bottom >= top???\n");

  newcell = bottom ;
  bottom += 2 ;
  __RPLACA(newcell,__CAR(p)) ;
  __RPLACD(newcell,__CDR(p)) ;
  return newcell;
}




/* BASIC LIST OPERATIONS */

inline int atom(MEM x)      { return IS_ATOM(x) || !in_tospace(M2P(x)) ; }
inline int eq(MEM x, MEM y) { return (x == y); }

inline PTR cons(MEM car, MEM cdr) {
  PTR newcell;

  DBG_printf("cons: car=0x%08lX, cdr=0x08lX\n", car, cdr);

  if(flipping){
    MOVE(car);
    MOVE(cdr);
  }

  GC_increment();

  if(bottom == top) error("cons: memory is full!!!\n");
  /* then k must be increased a lot,
     a full non-real-time GC must be completed,
     etc.     
   */

  newcell=bottom;
  bottom+=2;
  DBG_printf("cons: newcell=%p\n",newcell);
  __RPLACA(newcell,car);
  __RPLACD(newcell,cdr);
  DBG_printf("cons: newcell=%p\n",newcell);
  return newcell;
}

inline MEM car(PTR p) { return MOVE(__CAR(p)); }
inline MEM cdr(PTR p) { return MOVE(__CDR(p)); }

inline void rplaca(PTR x, MEM y) { __RPLACA(x,y); }
inline void rplacd(PTR x, MEM y) { __RPLACD(x,y); }
